defmodule FrenzyWeb.FeedController do
  use FrenzyWeb, :controller
  alias Frenzy.{Repo, Group, Feed, Item, Pipeline}
  alias FrenzyWeb.Router.Helpers, as: Routes
  alias FrenzyWeb.Endpoint
  import Ecto.Query

  plug :user_owns_feed

  defp user_owns_feed(%Plug.Conn{path_params: %{"id" => id}} = conn, _opts) do
    user = conn.assigns[:user]

    feed = Repo.get(Feed, id)

    if Enum.any?(user.groups, fn g -> g.id == feed.group_id end) do
      conn
      |> assign(:feed, feed)
    else
      conn
      |> put_flash(:error, "You do not have permission to access that resource.")
      |> redirect(to: Routes.group_path(Endpoint, :index))
      |> halt()
    end
  end

  defp user_owns_feed(conn, _opts), do: conn

  def show(conn, %{"id" => id}) do
    feed = conn.assigns[:feed]
    items = Repo.all(from Item, where: [feed_id: ^id, tombstone: false], order_by: [desc: :date])

    render(conn, "show.html", %{
      feed: feed,
      items: items
    })
  end

  def create(conn, %{"feed" => %{"group_id" => group_id, "feed_url" => feed_url}}) do
    user = conn.assigns[:user]
    group = Repo.get(Group, group_id)
    feed_url = feed_url |> String.trim()

    cond do
      group != nil && group.user_id == user.id ->
        if String.length(feed_url) > 0 do
          changeset =
            Ecto.build_assoc(group, :feeds, %{
              feed_url: feed_url
            })

          {:ok, _feed} = Repo.insert(changeset)

          conn
          |> put_flash(:info, "Feed created")
          |> redirect(to: Routes.group_path(Endpoint, :show, group_id))
        else
          conn
          |> put_flash(:error, "Feed URL must not be empty")
          |> redirect(to: Routes.group_path(Endpoint, :show, group_id))
        end

      true ->
        conn
        |> put_flash(:error, "Invalid group")
        |> redirect(to: Routes.group_path(Endpoint, :index))
    end
  end

  def edit(conn, _params) do
    feed = conn.assigns[:feed]
    edit_changeset = Feed.changeset(feed, %{})

    pipelines =
      Repo.all(Pipeline)
      |> Enum.map(fn pipeline -> {pipeline.name, pipeline.id} end)

    render(conn, "edit.html", %{
      feed: feed,
      changeset: edit_changeset,
      pipelines: [{"No Pipeline", nil} | pipelines]
    })
  end

  def update(conn, %{"feed" => feed_params}) do
    feed = conn.assigns[:feed]
    changeset = Feed.changeset(feed, feed_params)
    {:ok, feed} = Repo.update(changeset)
    redirect(conn, to: Routes.feed_path(Endpoint, :show, feed.id))
  end

  def delete(conn, _params) do
    feed = conn.assigns[:feed]
    {:ok, _} = Repo.delete(feed)
    redirect(conn, to: Routes.group_path(Endpoint, :show, feed.group_id))
  end

  def enable_filter(conn, _params) do
    feed = conn.assigns[:feed]
    changeset = Feed.changeset(feed, %{filter_enabled: true})
    Repo.update(changeset)
    redirect(conn, to: Routes.feed_path(Endpoint, :show, feed.id))
  end

  def disable_filter(conn, _params) do
    feed = conn.assigns[:feed]
    changeset = Feed.changeset(feed, %{filter_enabled: false})
    Repo.update(changeset)
    redirect(conn, to: Routes.feed_path(Endpoint, :show, feed.id))
  end

  def refresh(conn, _params) do
    feed = conn.assigns[:feed]
    feed = Frenzy.UpdateFeeds.refresh(Frenzy.UpdateFeeds, feed)

    conn
    |> put_flash(:info, "Refreshing feed. Wait a moment before reloading...")
    |> redirect(to: Routes.feed_path(Endpoint, :show, feed.id))
  end
end
